/*
 * Copyright (c) 2006 Hypertriton, Inc. <http://www.hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/stat.h>
#include <sys/param.h>
#include <sys/wait.h>

#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>

#include "mailprocd.h"
#include "pathnames.h"
	
static const char *mlOps[] = {
	"admin", "bounces", "confirm", "join", "leave", "owner", "request",
	"subscribe", "unsubscribe",
	NULL
};

int
ML_Init(void)
{
	return (0);
}

void
ML_Destroy(void)
{
}

/*
 * Search for a mailing list address matching the given recipient.
 * Returns the owner of the list of owner is not NULL. The listName
 * buffer must be at least ML_NAME_MAX bytes.
 */
ML_MailListReq *
ML_GetMailListReq(char *rcpt_addr)
{
	char list_addr[ML_NAME_MAX+16];
	char user_part[ADDRESS_MAX];
	char domain_part[ADDRESS_MAX];
	ML_MailListReq *ml;
	const char **op;
	char *owner_name, *maxsize, *vp;
	struct passwd *owner;
	char *rcpt_op;

	if (SF_ParseRecipientParts(rcpt_addr, user_part, domain_part) == -1)
		return (NULL);

	/* Look for an exact match on the list address. */
	if ((vp = table_lookup(Tlists, rcpt_addr)) != NULL) {
		owner_name = strsep(&vp, ":");
		maxsize = strsep(&vp, ":");
		if ((owner = getpwnam(owner_name)) == NULL) {
			syslog(LOG_ERR, "Bad %s owner: %s", rcpt_addr,
			    owner_name);
			return (NULL);
		}
		ml = Malloc(sizeof(ML_MailListReq));
		ml->owner = owner;
		ml->maxSize = atoi(maxsize);
		if (ml->maxSize == 0) { ml->maxSize = ML_DEFAULT_MAXSIZE; }
		Strlcpy(ml->op, "post", sizeof(ml->op));
		Strlcpy(ml->name, user_part, sizeof(ml->name));
	
		Debug("Mailing list match (name=%s, op=POST, owner=%s(%d), "
		      "maxpost=%d)\n",
		    ml->name, ml->owner->pw_name, ml->owner->pw_uid,
		    (int)ml->maxSize);
		return (ml);
	}

	/* Look for a mailing list operation address. */
	if ((rcpt_op = strrchr(user_part, '-')) == NULL ||
	    (rcpt_op[1] == '\0')) {
		return (NULL);
	}
	for (op = mlOps; *op != NULL; op++) {
		if (strcmp(&rcpt_op[1], *op) == 0)
			break;
	}
	if (*op == NULL) {
		return (NULL);
	}
	*rcpt_op = '\0';
	Strlcpy(list_addr, user_part, sizeof(list_addr));
	Strlcat(list_addr, "@", sizeof(list_addr));
	Strlcat(list_addr, domain_part, sizeof(list_addr));

	if ((vp = table_lookup(Tlists, list_addr)) == NULL) {
		return (NULL);
	}
	owner_name = strsep(&vp, ":");
	maxsize = strsep(&vp, ":");
	if ((owner = getpwnam(owner_name)) == NULL) {
		syslog(LOG_ERR, "Bad %s owner: %s", rcpt_addr, owner_name);
		return (NULL);
	}
	ml = Malloc(sizeof(ML_MailListReq));
	ml->owner = owner;
	ml->maxSize = atoi(maxsize);
	if (ml->maxSize == 0) { ml->maxSize = ML_DEFAULT_MAXSIZE; }
	Strlcpy(ml->op, *op, sizeof(ml->op));
	Strlcpy(ml->name, user_part, sizeof(ml->name));

	Debug("Mailing list match (name=%s, op=%s, owner=%s(%d), maxpost=%d)\n",
	    ml->name, ml->op, ml->owner->pw_name, ml->owner->pw_uid,
	    (int)ml->maxSize);
	return (ml);
}

/* Mailing list message delivery routine. */
void
ML_MessageProcess(SF_Message *msg)
{
	char cmd[sizeof(_PATH_MAILMAN)+ML_NAME_MAX+16];
	struct group *group;
	uid_t uid;
	gid_t gid;
	sigset_t allsigs;

	/*
	 * Drop privileges irrevocably since we don't have to handle
	 * multiple recipients. Use the _mailman group so the mailman
	 * executable can be invoked.
	 */
	if (LOCAL_GetDefaultRecipientUID(msg->rcpt_to, &uid, &gid) == -1) {
		SF_SetError("List(%s): Cannot get default UID");
		goto fail_temp;
	}
	if ((group = getgrnam("_mailman")) == NULL) {
		SF_SetError("No _mailman group");
		goto fail_temp;
	}
	gid = group->gr_gid;
	sigfillset(&allsigs);
	sigprocmask(SIG_BLOCK, &allsigs, NULL);
	if (setegid(gid) < 0 || setgid(gid) < 0) {
		SF_SetError("setgid(%d): %s", gid, strerror(errno));
		sigprocmask(SIG_UNBLOCK, &allsigs, NULL);
		goto fail_temp;
	}
	if (seteuid(uid) < 0 || setuid(uid) < 0) {
		SF_SetError("setuid(%d): %s", uid, strerror(errno));
		sigprocmask(SIG_UNBLOCK, &allsigs, NULL);
		goto fail_temp;
	}
	sigprocmask(SIG_UNBLOCK, &allsigs, NULL);

#ifdef HAVE_SA
	/*
	* Always run the message through SpamAssassin if this is
	* a "post" operation.
	*/
	if (strcmp(msg->ml->op, "post") == 0) {
		if (SA_ParseMessage(msg) == -1) {
			syslog(LOG_ERR, "SA_ParseMessage failed");
			goto sa_out;
		}
		if (SA_CheckMessage(msg) == -1) {
			syslog(LOG_ERR, "SA_CheckMessage failed");
			goto sa_out;
		}
		if (SA_RewriteMessage(msg) == -1) {
			syslog(LOG_ERR, "SA_RewriteMessage failed");
		}
		Debug("[%d] SpamAssassin score = %f", getpid(),
		    msg->status.score);
	}
sa_out:
#endif /* HAVE_SA */
	if (msg->status.score > 10.0) {
		Debug("Rejecting spam from <%s> to list %s",
		    msg->mail_from, msg->ml->name);
		return;
	}
	Debug("List <%s> operation %s (owner=%s)", msg->ml->name, msg->ml->op,
	    msg->ml->owner->pw_name);
	Strlcpy(cmd, _PATH_MAILMAN, sizeof(cmd));
	Strlcat(cmd, " ", sizeof(cmd));
	Strlcat(cmd, msg->ml->op, sizeof(cmd));
	Strlcat(cmd, " ", sizeof(cmd));
	Strlcat(cmd, msg->ml->name, sizeof(cmd));
	if (LOCAL_FeedToPipe(msg, cmd, msg->ml->op, msg->ml->owner->pw_uid,
	    gid) == -1) {
		goto fail_temp;
	}
	return;
fail_temp:
	Debug("List processor failed (%s); reinjecting", SF_GetError());
	sleep(3);
	if (SMTP_ForwardMessage(msg, sfReinHost, sfReinPort, msg->rcpt_to)
	    == -1) {
		syslog(LOG_ERR, "Reinjection failed (%s); queueing!",
		    SF_GetError());
		/* TODO QUEUE */
	}
}
