/*
 * Copyright (c) 2006 Hypertriton, Inc. <http://www.hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config/sharedir.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>

#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <netdb.h>
#include <pwd.h>
#include <unistd.h>

#include "config/version.h"
#include "mailprocd.h"
#include "pathnames.h"
#include "ports.h"

int sfPollInterval = 600;	/* Interval of maintenance ops (secs) */
int sfPolicyServer = 1;		/* Enable the Postfix policy server */
int sfCtlServer = 1;		/* Enable the control interface socket */

char *sfHostSMTP = "localhost";
char *sfPortSMTP = _PORT_MAILPROCD;
char *sfReinHost = "localhost";
char *sfReinPort = _PORT_REINJECT;
FILE *sfLog = NULL;

#ifdef HAVE_PERL
static PerlInterpreter *sfPerl;
#endif
static char sfError[128];
static int smtpd_socks[4], polSock, ctlSock;
static int nsmtpd_socks;
static volatile int die_flag = 0;
static volatile int reload_flag = 0;

void
SF_SetError(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vsnprintf(sfError, sizeof(sfError), fmt, ap);
	va_end(ap);
}
char *
SF_GetError(void)
{
	return (sfError);
}

void
SF_Fatal(const char *fmt, ...)
{
	char err[128];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(err, sizeof(err), fmt, ap);
	va_end(ap);
	fputs(err, stderr);
	fputc('\n', stderr);
	exit(1);
}

void *
SF_Malloc(size_t size)
{
	void *p;

	if ((p = malloc(size)) == NULL) {
		Fatal("Out of memory");
	}
	return (p);
}

#ifdef DEBUG
void
SF_Debug(const char *fmt, ...)
{
	static char msg[160];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);
	fprintf(sfLog, "[%d] %s\n", getpid(), msg);
	fflush(sfLog);
}
#endif /* DEBUG */

#ifdef HAVE_PERL
static void xs_init (pTHX);
EXTERN_C void boot_DynaLoader(pTHX_ CV *cv);
EXTERN_C void
xs_init(pTHX)
{
	char *file = __FILE__;
	dXSUB_SYS;
	newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
}
#endif /* HAVE_PERL */

static void
ReportChild(const char *what, pid_t rpid, int code)
{
	if (WIFSIGNALED(code)) {
		fprintf(sfLog, "[%d] %s exited (signal %d)\n",
		    (int)rpid, what, WTERMSIG(code));
	} else if (WIFEXITED(code) && WEXITSTATUS(code) != 0) {
		fprintf(sfLog, "[%d] %s exited due to failure (%d)\n",
		    (int)rpid, what, WEXITSTATUS(code));
	}
}

static void
sig_chld(int sigraised)
{
	int save_errno = errno;
	int rpid, code;

	do {
		rpid = waitpid(-1, &code, WNOHANG);
		if (polProc != 0 && rpid == polProc) {
			ReportChild("policy", rpid, code);
			polProc = 0;
		} else if (qmgrProc != 0 && rpid == qmgrProc) {
			ReportChild("qmgr", rpid, code);
			qmgrProc = 0;
		} else if (rpid > 0) {
			CTL_Session *sess;

			TAILQ_FOREACH(sess, &ctlSessions, sessions) {
				if (rpid == sess->pid)
					break;
			}
			if (sess != NULL) {
				TAILQ_REMOVE(&ctlSessions, sess, sessions);
				free(sess);
				ctlSessionCount--;
				Debug("Terminated control session: %d",
				    (int)sess->pid);
				ReportChild("session", rpid, code);
			} else {
				ReportChild("mailprocd", rpid, code);
			}
		}
	} while (rpid > 0 || (rpid == -1 && errno == EINTR));

	errno = save_errno;
}

static void
sig_quit(int sigraised)
{
#if 0
	syslog(LOG_ERR, "got signal %s", sys_signame[sigraised]);
#endif
	_exit(0);
}

static void
sig_die_server(int sigraised)
{
#if 0
	syslog(LOG_NOTICE, "server got signal %s",
	    sys_signame[sigraised]);
#endif
	die_flag = 1;
}

static void
sig_reload(int sigraised)
{
	reload_flag = 1;
}

static void
sig_pipe(int sigraised)
{
#if 0
	syslog(LOG_NOTICE, "broken pipe");
#endif
	_exit(0);
}

void
SF_EnterServerProc(const char *name)
{
	int i;
	struct sigaction sa;

	for (i = 0; i < nsmtpd_socks; i++) { close(smtpd_socks[i]); }
	if (sfPolicyServer) { close(polSock); }
	if (sfCtlServer) { close(ctlSock); }

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = SIG_DFL;
	sigaction(SIGCHLD, &sa, NULL);
	sa.sa_handler = SIG_IGN;
	sigaction(SIGURG, &sa, NULL);
	sigaction(SIGHUP, &sa, NULL);
	sa.sa_handler = sig_pipe;
	sigaction(SIGPIPE, &sa, NULL);
	
	sigfillset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sa.sa_handler = sig_quit;
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGQUIT, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGUSR1, &sa, NULL);

	closelog();
	openlog(name, LOG_PID|LOG_NDELAY, LOG_LOCAL0);
	Setproctitle("%s", name);
}

void
SF_ExitServerProc(int code)
{
	closelog();
	exit(code);
}

static void
SF_OpenLog(void)
{
	int fd;

	if (sfLog != NULL) {
		fclose(sfLog);
	}
	if ((sfLog = fopen(_PATH_LOG_FILE, "a")) == NULL) {
		syslog(LOG_ERR, "%s: %s", _PATH_LOG_FILE, strerror(errno));
		fprintf(stderr, "%s: %s\n", _PATH_LOG_FILE, strerror(errno));
		exit(1);
	}
	fd = fileno(sfLog);
	fchmod(fd, 0640);
	fchown(fd, 0, 20);
}

int
main(int argc, char **argv, char **env)
{
#ifdef HAVE_PERL
	char initPath[1024];
	char *myArgv[2] = { "", initPath };
#endif
	struct addrinfo hints, *res, *res0;
	const char *cause = NULL;
	int rv, i, maxfd = 0;
	fd_set servfds;
	struct sigaction sa;
	my_socklen_t socklen;
	struct stat sb;
	FILE *f;
	struct sockaddr_un unaddr, paddr;
	struct passwd *pwd;

#ifdef HAVE_PERL
	Strlcpy(initPath, SHAREDIR, sizeof(initPath));
	Strlcat(initPath, "/init.pl", sizeof(initPath));
	PERL_SYS_INIT3(&argc, &argv, &env);
	sfPerl = perl_alloc();
	perl_construct(sfPerl);
	perl_parse(sfPerl, xs_init, 2, myArgv, (char **)NULL);
	PL_exit_flags |= PERL_EXIT_DESTRUCT_END;
	perl_run(sfPerl);
#endif

	SF_OpenLog();

#ifdef HAVE_SA
	if (SA_GetVersion(saVersion, sizeof(saVersion)) == -1) {
		Fatal("Cannot get SA version");
	}
	Debug("mailprocd %s (SpamAssassin %s)", VERSION, saVersion);
	syslog(LOG_INFO, "Server startup (%s with SpamAssassin %s)", VERSION,
	    saVersion);
	Debug("Compiling SpamAssassin patterns...");
	SA_CompileNow(0, 0);
	Debug("OK");
#else
	Debug("mailprocd %s", VERSION);
	syslog(LOG_INFO, "Server startup (%s)", VERSION);
#endif

	if (stat(_PATH_PID, &sb) == 0) {
		Debug("WARNING: Overwriting existing PID file!");
		syslog(LOG_WARNING, "Existing PID file!");
	}
	if ((f = fopen(_PATH_PID, "w")) == NULL) {
		Debug("%s: %s", _PATH_PID, strerror(errno));
		syslog(LOG_ERR, "%s: %s", _PATH_PID, strerror(errno));
		exit(1);
	}
	fprintf(f, "%lu\n", (unsigned long)getpid());
	fclose(f);

	/*
	 * Initialization of the subsystems
	 */
	if (stat(_PATH_HOME, &sb) != 0 &&
	    mkdir(_PATH_HOME, 0755) == -1) {
		SF_SetError("mkdir %s: %s", _PATH_HOME, strerror(errno));
		goto fatal;
	}
	if (stat(_PATH_SOCKETDIR, &sb) != 0 &&
	    mkdir(_PATH_SOCKETDIR, 0755) == -1) {
		SF_SetError("mkdir %s: %s", _PATH_SOCKETDIR, strerror(errno));
		goto fatal;
	}
	if (stat(_PATH_SAUSERDIR, &sb) != 0 &&
	    mkdir(_PATH_SAUSERDIR, 0755) == -1) {
		SF_SetError("mkdir %s: %s", _PATH_SAUSERDIR, strerror(errno));
		goto fatal;
	}
	if (LOCAL_Init() == -1) {
		goto fatal;
	}
	if (QMGR_Init() == -1) {
		LOCAL_Destroy();
		goto fatal;
	}
	if (POL_Init() == -1) {
		LOCAL_Destroy();
		goto fatal;
	}
	if (CTL_Init() == -1) {
		POL_Destroy();
		LOCAL_Destroy();
		goto fatal;
	}

	/*
	 * Set up the SMTP server listening socket(s).
	 */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	if ((rv = getaddrinfo(sfHostSMTP, sfPortSMTP, &hints, &res0)) != 0) {
		syslog(LOG_ERR, "%s:%s: %s", sfHostSMTP, sfPortSMTP,
		    gai_strerror(rv));
		goto out;
	}
	FD_ZERO(&servfds);
	for (nsmtpd_socks = 0, res = res0;
	     res != NULL && nsmtpd_socks < 4;
	     res = res->ai_next) {
		rv = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (rv == -1) {
			cause = "socket";
			continue;
		}
		i = 1;
		setsockopt(rv, SOL_SOCKET, SO_REUSEADDR, &i,
		    (my_socklen_t)sizeof(i));
		if (bind(rv, res->ai_addr, res->ai_addrlen) == -1) {
			cause = "bind";
			close(rv);
			continue;
		}
		if (listen(rv, 10) == -1) {
			cause = "listen";
			close(rv);
			continue;
		}
		smtpd_socks[nsmtpd_socks++] = rv;
		FD_SET(rv, &servfds);
		if (rv > maxfd) { maxfd = rv; }
	}
	if (nsmtpd_socks == 0) {
		syslog(LOG_ERR, "%s: %s", cause, strerror(errno));
		goto out;
	}
	freeaddrinfo(res0);

	/*
	 * Set up the Control Interface socket.
	 */
	if (sfCtlServer) {
		if ((ctlSock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
			syslog(LOG_ERR, "socket(AF_UNIX): %s", strerror(errno));
			goto out;
		}
		unaddr.sun_family = AF_UNIX;
		Strlcpy(unaddr.sun_path, ctlSockPath, sizeof(unaddr.sun_path));
		unlink(unaddr.sun_path);
		socklen = SUN_LEN(&unaddr);
		if (bind(ctlSock, (struct sockaddr *)&unaddr, socklen) == -1 ||
		    listen(ctlSock, 5) == -1) {
			syslog(LOG_ERR, "%s: %s", unaddr.sun_path,
			    strerror(errno));
			goto out;
		}
		chmod(unaddr.sun_path, 0777);
		FD_SET(ctlSock, &servfds);
		if (ctlSock > maxfd) { maxfd = ctlSock; }
	}

	/*
	 * Set up the Policy Server socket.
	 */
	if (sfPolicyServer) {
		if ((polSock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
			syslog(LOG_ERR, "socket(AF_UNIX): %s", strerror(errno));
			goto out;
		}
		unaddr.sun_family = AF_UNIX;
		Strlcpy(unaddr.sun_path, polSockPath, sizeof(unaddr.sun_path));
		unlink(unaddr.sun_path);
		socklen = SUN_LEN(&unaddr);
		if (bind(polSock, (struct sockaddr *)&unaddr, socklen) == -1 ||
		    listen(polSock, 5) == -1) {
			syslog(LOG_ERR, "%s: %s", unaddr.sun_path,
			    strerror(errno));
			goto out;
		}
		if ((pwd = getpwnam(polOwner)) != NULL) {
			chown(unaddr.sun_path, pwd->pw_uid, pwd->pw_gid);
			chmod(unaddr.sun_path, 0700);
		} else {
			syslog(LOG_ERR, "Invalid policy owner: %s", polOwner);
		}
		
		FD_SET(polSock, &servfds);
		if (polSock > maxfd) { maxfd = polSock; }
	}

	sigemptyset(&sa.sa_mask);

	sa.sa_handler = sig_chld;
	sa.sa_flags = SA_RESTART;
	sigaction(SIGCHLD, &sa, NULL);
	
	sa.sa_handler = sig_reload;
	sa.sa_flags = 0;
	sigaction(SIGHUP, &sa, NULL);
	
	sa.sa_handler = sig_die_server;
	sa.sa_flags = 0;
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGQUIT, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);

	sa.sa_handler = SIG_IGN;
	sigaction(SIGUSR1, &sa, NULL);
	
	for (;;) {
		fd_set rfds = servfds;
		struct timeval tv;

		tv.tv_sec = sfPollInterval;
		tv.tv_usec = 0;
		rv = select(maxfd+1, &rfds, NULL, NULL, &tv);
		if (rv == 0) {
			if (qmgrProc != 0 && QMGR_Todo() > 0) {
				QMGR_ForkInstance();
			}
		} else if (rv == -1) {
			if (errno == EINTR) {
				if (die_flag) {
					kill(0, SIGUSR1);
					goto out_close;
				}
				if (reload_flag) {
					reload_flag = 0;
					LOCAL_ReloadDatabases();
					SF_OpenLog();
				}
				continue;
			} else {
				syslog(LOG_ERR, "select: %s; shutdown",
				    strerror(errno));
				exit(1);
			}
		}
		if (die_flag) {
			kill(0, SIGUSR1);
			goto out_close;
		}
		if (reload_flag) {
			reload_flag = 0;
			LOCAL_ReloadDatabases();
		}

		if (sfCtlServer && FD_ISSET(ctlSock, &rfds)) {
			socklen = sizeof(paddr);
			rv = accept(ctlSock, (struct sockaddr *)&paddr,
			    &socklen);
			if (rv != -1) {
				CTL_ForkInstance(rv);
				close(rv);
			} else {
				syslog(LOG_ERR, "accept(control): %s",
				    strerror(errno));
			}
		}
		if (sfPolicyServer && FD_ISSET(polSock, &rfds)) {
			socklen = sizeof(paddr);
			rv = accept(polSock, (struct sockaddr *)&paddr,
			    &socklen);
			if (rv != -1) {
				POL_ForkInstance(rv);
				close(rv);
			} else {
				syslog(LOG_ERR, "accept(policy): %s",
				    strerror(errno));
			}
		}
		for (i = 0; i < nsmtpd_socks; i++) {
			union sockunion paddr;
			pid_t pid;

			if (!FD_ISSET(smtpd_socks[i], &rfds)) {
				continue;
			}
			socklen = sizeof(paddr);
			rv = accept(smtpd_socks[i], (struct sockaddr *)&paddr,
			    &socklen);
			if (rv == -1) {
				syslog(LOG_ERR, "accept(smtp): %s",
				    strerror(errno));
				continue;
			}
			if ((pid = fork()) == 0) {		/* Child */
				dup2(rv, 0);
				dup2(rv, 1);

				SF_EnterServerProc("mailprocd");
				i = 1;
				setsockopt(0, SOL_SOCKET, SO_OOBINLINE, &i,
				    sizeof(i));
				fcntl(fileno(stdin), F_SETOWN, getpid());
				if (setvbuf(stdout, NULL, _IOLBF, 0) == EOF) {
					syslog(LOG_ERR, "_IOLBF stdout: %m");
					exit(1);
				}
				if (SMTP_Session() == -1) {
					syslog(LOG_ERR, "SMTP_Session: %s",
					    SF_GetError());
				}
				SF_ExitServerProc(0);
			}
			close(rv);
		}
	}
out_close:
	syslog(LOG_NOTICE, "Server shutdown (signal)");
	for (i = 0; i < nsmtpd_socks; i++) {
		close(smtpd_socks[i]);
	}
	if (sfPolicyServer) { close(polSock); }
	if (sfCtlServer) { close(ctlSock); }
	CTL_Destroy();
	POL_Destroy();
	LOCAL_Destroy();
#ifdef HAVE_SA
	SPAM_Destroy();
#endif
	unlink(_PATH_PID);
out:
#ifdef HAVE_PERL
	perl_destruct(sfPerl);
	perl_free(sfPerl);
	PERL_SYS_TERM();
#endif
	return (0);
fatal:
	fprintf(stderr, "FATAL: %s\n", SF_GetError());
	syslog(LOG_CRIT, "FATAL: %s", SF_GetError());
	return (1);
}
