/*
 * Copyright (c) 2006-2007 Hypertriton, Inc. <http://www.hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <netdb.h>

#include "mailprocd.h"

static __inline__ void
SMTP_WriteClient(const char *s)
{
	fputs(s, stdout);
#if 0
	fflush(stdout);
#endif
}

static int
SMTP_Expect(FILE *f, const char *code)
{
	char *buf, *lbuf = NULL;
	size_t len;
	int rv = 0;

	if ((buf = Fgetln(f, &len)) == NULL) {
		SF_SetError("EOF from server");
		return (0);
	}
	if (buf[len-1] == '\n') {
		buf[len-1] = '\0';
	} else {
		if ((lbuf = malloc(len+1)) == NULL) {
			SF_SetError("malloc linebuf");
			return (0);
		}
		memcpy(lbuf, buf, len);
		lbuf[len] = '\0';
		buf = lbuf;
	}
	if (strncmp(buf, code, 3) != 0) {
		SF_SetError("returned \"%s\" (was expecting %s)", buf, code);
		rv = 0;
	}
	if (lbuf != NULL) { free(lbuf); }
	return (1);
}

/*
 * Forward a message back to the MTA, or place it in the queue if the
 * MTA is not responding.
 */
int
SMTP_ForwardMessage(SF_Message *msg, const char *rhost, const char *rport,
    const char *rcpt)
{
	struct addrinfo hints, *res, *res0;
	const char *cause = NULL;
	int rv, s;
	FILE *f;

	Debug("Forwarding message to <%s> via [%s:%s] (%lu bytes)", 
	    rcpt, rhost, rport, (unsigned long)msg->text_len);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	if ((rv = getaddrinfo(rhost, rport, &hints, &res0)) != 0) {
		SF_SetError("%s:%s: %s", rhost, rport, gai_strerror(rv));
		return (-1);
	}
	for (s = -1, res = res0;
	     res != NULL;
	     res = res->ai_next) {
		s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (s < 0) {
			cause = "socket";
			continue;
		}
		if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
			cause = "connect";
			close(s);
			s = -1;
			continue;
		}
		break;
	}
	if (s == -1) {
		SF_SetError("%s: %s", cause, strerror(errno));
		goto fail_addr;
	}
	if ((f = fdopen(s, "w+")) == NULL) {
		SF_SetError("fdopen: %s", strerror(errno));
		goto fail;
	}
	if (setvbuf(f, NULL, _IOLBF, 0) == EOF) {
		syslog(LOG_ERR, "setvbuf _IOLBF failed");
	}

	if (!SMTP_Expect(f, "220")) { goto fail; }
	fputs("HELO [127.0.0.1]\n", f);
	if (!SMTP_Expect(f, "250")) { goto fail; }
	fputs("MAIL FROM:<", f);
	fputs(msg->mail_from, f);
	fputs(">\n", f);
	if (!SMTP_Expect(f, "250")) { goto fail; }
	fputs("RCPT TO:<", f);
	fputs(rcpt, f);
	fputs(">\n", f);
	if (!SMTP_Expect(f, "250")) { goto fail; }
	fputs("DATA\n", f);
	if (!SMTP_Expect(f, "354")) { goto fail; }
	fputs(msg->text, f);
	fputs("\r\n.\r\n", f);
	if (!SMTP_Expect(f, "250")) { goto fail; }

	Debug("Reinjected message successfully");

	close(s);
	freeaddrinfo(res0);
	return (0);
fail:
	close(s);
fail_addr:
	freeaddrinfo(res0);
	return (-1);
}

/* Remove whitespace and brackets from an address. */
static __inline__ void
SMTP_ParseAddress(char *dest, char *buf, size_t dest_len)
{
	char *c, *cStart;

	for (c = &buf[0]; *c == ' ' || *c == '<'; c++)
		;;
	cStart = c;

	for (; *c != '>' && *c != '\0'; c++)
		;;
	if (*c == '>')
		*c = '\0';

	Strlcpy(dest, cStart, dest_len);
	for (c = &dest[0]; *c != '\0'; c++) {
		*c = tolower((int)*c);
	}
}

static void
goodbye(void)
{
	SMTP_WriteClient("221 localhost\n");
	shutdown(fileno(stdin), SHUT_RDWR);
	shutdown(fileno(stdout), SHUT_RDWR);
}

static int
quit_cmd(const char *s)
{
	if (strcasecmp(s, "quit") == 0) {
		goodbye();
		return (1);
	}
	return (0);
}

/* Process an incoming SMTP session initiated by the MTA. */
int
SMTP_Session(void)
{
	char *buf, *lbuf = NULL, *pbuf;
	size_t len;
	SF_Message *msg;
	enum {
		ST_HELO,
		ST_MAIL_FROM,
		ST_RCPT_TO,
		ST_DATA,
		ST_DATA_IN,
		ST_QUIT
	} state = ST_HELO;
	int in_header = 1;
	int looping = 0;

	SMTP_WriteClient("220 localhost ESMTP\n");

	msg = SF_MessageNew();
	if ((msg->text = malloc(1)) == NULL) {
		SF_SetError("malloc msg buf");
		return (-1);
	}
	msg->text[0] = '\0';
	msg->text_len = 0;

	while ((buf = Fgetln(stdin, &len)) != NULL) {
		if (buf[len-1] == '\n') {
			if (len > 1 && buf[len-2] == '\r') {
				len--;
			}
			buf[len-1] = '\0';
		} else {
			if ((lbuf = malloc(len+1)) == NULL) {
				SF_SetError("malloc line");
				goto fail;
			}
			memcpy(lbuf, buf, len);
			lbuf[len] = '\0';
			buf = lbuf;
		}
		switch (state) {
		case ST_HELO:
			if (strncasecmp(buf, "helo", 4) == 0) {
				SMTP_WriteClient("221 localhost\n");
				state = ST_MAIL_FROM;
			} else {
				if (quit_cmd(buf)) { goto out; }
				SMTP_WriteClient("502 helo first\n");
			}
			break;
		case ST_MAIL_FROM:
			if (strncasecmp(buf, "mail from:", 10) == 0) {
				SMTP_ParseAddress(msg->mail_from, &buf[10],
				    sizeof(msg->mail_from));
				SMTP_WriteClient("250 ok\n");
				state = ST_RCPT_TO;
			} else {
				if (quit_cmd(buf)) { goto out; }
				SMTP_WriteClient("502 from first\n");
			}
			break;
		case ST_RCPT_TO:
			if (strncasecmp(buf, "rcpt to:", 8) == 0) {
				SMTP_ParseAddress(msg->rcpt_to, &buf[8],
				    sizeof(msg->rcpt_to));
				if (SF_ParseRecipientParts(msg->rcpt_to,
				    msg->user_part, msg->domain_part) == -1) {
					SF_SetError("Malformed recipient");
					goto fail;
				}
				SMTP_WriteClient("250 ok\n");
				state = ST_DATA;
			} else {
				if (quit_cmd(buf)) { goto out; }
				SMTP_WriteClient("502 to first\n");
			}
			break;
		case ST_DATA:
			if (strncasecmp(buf, "data", 4) == 0) {
				SMTP_WriteClient("354 go ahead\n");
				state = ST_DATA_IN;
				in_header = 1;
			} else {
				//SMTP_WriteClient("502 data first\n");
				if (quit_cmd(buf)) { goto out; }
				SMTP_WriteClient("450 data first\n");
			}
			break;
		case ST_DATA_IN:
			if (buf[0] == '.' && buf[1] == '\0') {
				if (looping) {
					SMTP_WriteClient("550 Mail loop!\n");
					Debug("Mail loop!");
				} else {
					SMTP_WriteClient("250 ok\n");
					Debug("From <%s>, to <%s>, %lu bytes",
					    msg->mail_from, msg->rcpt_to,
					    (unsigned long)msg->text_len);
					msg->text[msg->text_len] = '\0';
				}
				state = ST_QUIT;
			} else {
				if (buf[0] == '\0') {
					in_header = 0;
				} else if (!in_header && buf[0] == 'F' &&
				    strncmp(buf, "From", 4) == 0) {
					/*
					 * Quote /^From/ lines for future
					 * mailbox delivery (it is more
					 * efficient to do it here).
					 */
					pbuf = realloc(msg->text,
					               msg->text_len+1);
					if (pbuf == NULL) {
						SF_SetError("Out of memory");
						goto fail;
					}
					msg->text = pbuf;
					msg->text[msg->text_len] = '>';
					msg->text_len++;
				} else if (in_header && buf[0] == 'X' &&
				    strncmp(buf, "X-Csoft-Pipe: ", 14) == 0) {
					if (strcasecmp(&buf[14], msg->rcpt_to)
					    == 0) {
						looping++;
					}
				}
				pbuf = realloc(msg->text, msg->text_len+len+1);
				if (pbuf == NULL) {
					SF_SetError("Out of memory");
					goto fail;
				}
				msg->text = pbuf;
				buf[len-1] = '\n';
				memcpy(&msg->text[msg->text_len], buf, len);
				msg->text_len += len;
			}
			break;
		case ST_QUIT:
			if (strncasecmp(buf, "quit", 4) == 0) {
				goodbye();
				goto out;
			} else if (strncasecmp(buf, "rset", 4) == 0) {
				Debug("RSET not implemented!");
				SMTP_WriteClient("502 unimplemented\n");
			}
#if 0
				/* XXX privileges have already been dropped */
				SMTP_WriteClient("220 flushed\n");
				state = ST_MAIL_FROM;
#endif
			break;
		default:
			Debug("SMTP unimplemented: %s\n", buf);
			SMTP_WriteClient("502 unimplemented\n");
			break;
		}
	}
	if (ferror(stdin)) {
		SF_SetError("read error");
		goto fail;
	} else if (feof(stdin)) {
		SF_SetError("EOF");
		goto fail;
	}
out:
	if (looping) {
		SF_SetError("Mail loop generated by user command");
		goto fail;
	}
	if (lbuf != NULL) {
		free(lbuf);
	}

	if ((msg->ml = ML_GetMailListReq(msg->rcpt_to)) != NULL) {
		ML_MessageProcess(msg);
	} else {
		SF_MessageProcess(msg);
	}
	SF_MessageFree(msg);
	return (0);
fail:
	if (lbuf != NULL) { free(lbuf); }
	SF_MessageFree(msg);
	return (-1);
}

