/*
 * Copyright (c) 2007 Hypertriton, Inc. <http://hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_SA

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "mailprocd.h"

#include <stdlib.h>
#include <limits.h>
#include <syslog.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>

#include "pathnames.h"

/* Die after processing this number of messages */
static const int MaxMsgsPerProc = 100;

/* Maximum idle time in seconds for the process. */
static const int MaxIdle = 5*60;

static volatile int die_flag = 0;

static int
CreateDefaultDir(const char *saPath)
{
	if (mkdir(saPath, 0700) == -1) {
		SF_SetError("Failed to create %s: %s", saPath, strerror(errno));
		return (-1);
	}
	if (chdir(saPath) == -1) {
		SF_SetError("chdir %s: %s", saPath, strerror(errno));
		return (-1);
	}
	return (0);
}

static void
ProcessSetupOK(int fdParent)
{
	const char code = '0';

	if (write(fdParent, &code, 1) == -1) {
		syslog(LOG_ERR, "write(0) to parentfd: %s", strerror(errno));
		exit(1);
	}
}

static void
ProcessSetupFailed(int fdParent)
{
	const char code = '1';

	syslog(LOG_ERR, "spamcheck(%ld): setup failed: %s\n", (long)getpid(),
	    SF_GetError());
	if (write(fdParent, &code, 1) == -1) {
		fprintf(stderr, "write(1) to parentfd: %s", strerror(errno));
		exit(1);
	}
}

static void
sig_die(int sigraised)
{
	die_flag = 1;
}

static int
CreateSpamCheckProcess(SF_Message *msg, const char *sockPath, struct passwd *pw)
{
	extern char **environ;
	char saConfigPath[1024];
	char saPath[1024];
	char code;
	int pp[2];
	char wbuf[32];
	char *cleanenv[1];
	struct sigaction sa;
	sigset_t allsigs;
	size_t wrote;
	struct sockaddr_un unaddr;
	my_socklen_t socklen;
	int servSock;
	pid_t pid;
	unsigned int processedMsgs = 0;
	fd_set servfds;
	
	if (pipe(pp) == -1) {
		SF_SetError("pipe: %s", strerror(errno));
		return (-1);
	}
	if ((pid = fork()) > 0) {
		close(pp[1]);

		/* Block until child is ready to process requests. */
		if (read(pp[0], &code, 1) != 1) {
			SF_SetError("Failed to read status from spamcheck");
			return (-1);
		}
		if (code != '0') {
			SF_SetError("spamcheck process returned error status");
			return (-1);
		}
		close(pp[0]);
		return (0);
	} else if (pid == -1) {
		SF_SetError("fork: %s", strerror(errno));
		return (-1);
	}

	/* Set up the spamcheck process. */
	Setproctitle("spamcheck");
	openlog("spamcheck", LOG_PID, LOG_LOCAL0);
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = SIG_DFL;
	sigaction(SIGPIPE, &sa, NULL);
	sigfillset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = sig_die;
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGQUIT, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGUSR1, &sa, NULL);
	close(pp[0]);
	close(STDIN_FILENO);
	
	/* Create our Unix socket. */
	Debug("Creating unix socket on %s", sockPath);
	if ((servSock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
		SF_SetError("socket(AF_UNIX): %s", strerror(errno));
		ProcessSetupFailed(pp[1]);
		exit(1);
	}
	unaddr.sun_family = AF_UNIX;
	Strlcpy(unaddr.sun_path, sockPath, sizeof(unaddr.sun_path));
	socklen = SUN_LEN(&unaddr);
	unlink(sockPath);
	if (bind(servSock, (struct sockaddr *)&unaddr, socklen) == -1 ||
	    listen(servSock, 5) == -1) {
		SF_SetError("%s: %s", sockPath, strerror(errno));
		ProcessSetupFailed(pp[1]);
		exit(1);
	}
	chown(sockPath, pw->pw_uid, pw->pw_gid);
	chmod(sockPath, 0700);
	FD_ZERO(&servfds);
	FD_SET(servSock, &servfds);
	
	/* Create the SpamAssassin database directory if needed */
	Strlcpy(saPath, _PATH_SAUSERDIR, sizeof(saPath));
	Strlcat(saPath, pw->pw_name, sizeof(saPath));
	if (chdir(saPath) == -1) {
		if (CreateDefaultDir(saPath) == -1) {
			ProcessSetupFailed(pp[1]);
			exit(1);
		}
	}
	chown(saPath, pw->pw_uid, pw->pw_gid);
	chmod(saPath, 0700);

	/* Clear the environment and drop privileges. */
	sigfillset(&allsigs);
	sigprocmask(SIG_BLOCK, &allsigs, NULL);
	environ = cleanenv;
	cleanenv[0] = NULL;
	setenv("USER", pw->pw_name, 1);
	setenv("LOGNAME", pw->pw_name, 1);
	setenv("HOME", pw->pw_dir, 1);
	setenv("SHELL", pw->pw_shell, 1);
	setenv("PATH", _SAFE_PATH, 1);
	if (setegid(pw->pw_gid) < 0 || setgid(pw->pw_gid) < 0 ||
	    seteuid(pw->pw_uid) < 0 || setuid(pw->pw_uid) < 0 ||
	    geteuid() == 0 || getuid() == 0 ||			/* Paranoia */
	    getegid() == 0 || getgid() == 0) {
		sigprocmask(SIG_UNBLOCK, &allsigs, NULL);
		SF_SetError("set*id(%d,%d): %s", pw->pw_uid, pw->pw_gid,
		    strerror(errno));
		ProcessSetupFailed(pp[1]);
		exit(1);
	}
	sigprocmask(SIG_UNBLOCK, &allsigs, NULL);
	
	/* Create ~/.spamassassin if it does not exist yet. */
	Strlcpy(saConfigPath, pw->pw_dir, sizeof(saConfigPath));
	Strlcat(saConfigPath, _PATH_SACONF, sizeof(saConfigPath));
	if (chdir(saConfigPath) == -1) {
		if (CreateDefaultDir(saConfigPath) == -1) {
			ProcessSetupFailed(pp[1]);
			exit(1);
		}
	}
	Strlcat(saConfigPath, _PATH_SAUSERPREFS, sizeof(saConfigPath));

	/* Set up SpamAssassin for this user. */
	SA_SignalUserChange(pw->pw_name, pw->pw_dir, saPath);
	SA_ReadScoreOnlyConfig(saConfigPath);

	/* Initialize the learning facility */
	if (SA_InitLearner() == -1) {
		SF_SetError("SA_InitLearner");
		ProcessSetupFailed(pp[1]);
		exit(1);
	}

	/* Free the message structure inherited from the parent. */
	SF_MessageFree(msg);

	/*
	 * Send an OK message back to the parent, close the pipe and
	 * start listening for requests on the socket.
	 */
	ProcessSetupOK(pp[1]);
	close(pp[1]);
	processedMsgs = 0;
	while (!die_flag) {
		fd_set rfds;
		struct sockaddr_un paddr;
		struct timeval tv;
		ssize_t nread;
		size_t len;
		int rv, sock;
		char *ep;

		Setproctitle("-spamcheck: %u msgs", processedMsgs++);
		if (MaxMsgsPerProc != 0 &&
		    processedMsgs > MaxMsgsPerProc) {
			Debug("Processed %d messages, exiting", processedMsgs);
			goto out;
		}

		socklen = sizeof(paddr);
tryaccept:
		rfds = servfds;
		tv.tv_sec = MaxIdle;
		tv.tv_usec = 0;
		rv = select(servSock+1, &rfds, NULL, NULL, &tv);
		if (rv == 0) {
			Debug("Idle for >%d seconds, exiting", MaxIdle);
			goto out;
		} else if (rv == -1) {
			if (errno == EINTR) {
				if (die_flag)	{ goto out; }
				else		{ goto tryaccept; }
			} else {
				SF_SetError("select: %s", strerror(errno));
				goto fail;
			}
		}

		sock = accept(servSock, (struct sockaddr *)&paddr, &socklen);
		if (sock == -1) {
			if (errno == EINTR) {
				if (die_flag)	{ goto out; }
				else		{ goto tryaccept; }
			} else {
				SF_SetError("accept: %s", strerror(errno));
				goto fail;
			}
		}
	
		/* Read new message from the socket. */
		msg = SF_MessageNew();
		if (read(sock, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
			SF_SetError("Bad len value");
			SF_MessageFree(msg);
			goto fail;
		}
		errno = 0;
		len = (size_t)strtoul(wbuf, &ep, 10);
		if (*ep != '\0' || errno == ERANGE) {
			SF_SetError("Bad len: %s", wbuf);
			SF_MessageFree(msg);
			goto fail;
		}
		msg->text_len = len;
		if ((msg->text = realloc(msg->text, len+1)) == NULL) {
			SF_SetError("Out of memory for rewritten message (%lu)",
			    (unsigned long)msg->text_len);
			msg->text_len = 0;
			SF_MessageFree(msg);
			goto fail;
		}
		for (nread = 0; nread < len;) {
			ssize_t rRd;

			rRd = read(sock, &msg->text[nread], len - nread);
			if (rRd == -1) {
				if (errno == EINTR) {
					continue;
				}
				SF_SetError("Read error: %s", strerror(errno));
				SF_MessageFree(msg);
				goto fail;
			} else if (rRd == 0) {
				syslog(LOG_ERR, "EOF from spamcheck process");
				break;
			}
			nread += rRd;
		}
		msg->text[nread] = '\0';

		/* Perform the SpamAssassin test and rewrite the message. */
		Debug("Analyzing message (%d bytes)", (int)nread);
		if (SA_ParseMessage(msg) == -1) {
			SF_SetError("SA_ParseMessage");
			SF_MessageFree(msg);
			goto fail;
		}
		if (SA_CheckMessage(msg) == -1) {
			SF_SetError("SA_CheckMessage");
			SF_MessageFree(msg);
			goto fail;
		}
		if (SA_RewriteMessage(msg) == -1) {
			SF_SetError("SA_RewriteMessage");
			SF_MessageFree(msg);
			goto fail;
		}
	
		/* Write the result score to the parent. */
		snprintf(wbuf, sizeof(wbuf), "%f", msg->status.score);
		if (write(sock, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
			SF_SetError("Write error (score)");
			SF_MessageFree(msg);
			goto fail;
		}
			
		/* Write the user's required score preference to the parent. */
		snprintf(wbuf, sizeof(wbuf), "%f", msg->status.req_score);
		if (write(sock, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
			SF_SetError("Write error (req_score)");
			SF_MessageFree(msg);
			goto fail;
		}
	
		/* Write the rewritten message back to the parent. */
		snprintf(wbuf, sizeof(wbuf), "%lu",
		    (unsigned long)msg->text_len);
		if (write(sock, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
			SF_SetError("Write error (len)");
			SF_MessageFree(msg);
			goto fail;
		}
		for (wrote = 0; wrote < msg->text_len;) {
			ssize_t rWr;

			rWr = write(sock, &msg->text[wrote],
			    msg->text_len - wrote);
			if (rWr == -1) {
				if (errno == EINTR) {
					continue;
				}
				SF_SetError("Write error: %s", strerror(errno));
				SF_MessageFree(msg);
				goto fail;
			} else if (rWr == 0) {
				SF_SetError("Parent EOF");
				SF_MessageFree(msg);
				goto fail;
			}
			wrote += rWr;
		}

		/* Learn from the message. */
		SA_LearnMessage(msg);
		SA_RebuildLearnerCaches();
		SF_MessageFree(msg);

		close(sock);
	}
out:
	/* Clean up */
	if (CTL_SendCommand(CTL_SPAMCHECK_EXIT) == -1) {
		Debug("Sending CTL_SPAMCHECK_EXIT: %s", SF_GetError());
	}
	SA_FinishLearner();
	SA_FinishAddrListFactory();
	SA_Finish();
	close(servSock);
	closelog();
	exit(0);
fail:
	if (CTL_SendCommand(CTL_SPAMCHECK_EXIT) == -1) {
		Debug("Sending CTL_SPAMCHECK_EXIT: %s", SF_GetError());
	}
	SA_FinishLearner();
	Debug("spamcheck failed (%s)", SF_GetError());
	syslog(LOG_ERR, "spamcheck failed: %s", SF_GetError());
	SA_FinishAddrListFactory();
	SA_Finish();
	close(servSock);
	closelog();
	exit(1);
}

int
SPAM_Check(SF_Message *msg)
{
	char sockPath[1024];
	char wbuf[32];
	int sock;
	uid_t uid;
	gid_t gid;
	char *ep;
	size_t len;
	ssize_t rv;
	size_t nread;
	ssize_t wrote;
	struct passwd *pw;
	struct sockaddr_un unaddr;
	my_socklen_t socklen;
	int retry = 0;

	if (msg->text_len >= SPAM_TEST_SIZE_LIMIT) {
		msg->status.score = 0.0;
		msg->status.req_score = 123.0;
		Debug("Message over limit for spam test (%lu bytes)",
		    (unsigned long)msg->text_len);
		return (0);
	}

	/* Find the UID/GID corresponding to the recipient address. */
	if (LOCAL_GetDefaultRecipientUID(msg->rcpt_to, &uid, &gid) == -1) {
		SF_SetError("Cannot get default UID for %s", msg->rcpt_to);
		return (-1);
	}
	if ((pw = getpwuid(uid)) == NULL) {
		SF_SetError("Bad UID for %s (%ld)", msg->rcpt_to, (long)uid);
		return (-1);
	}

	/*
	 * Create a spamcheck process if this user does not already have
	 * one running and set up the communication channel.
	 */
	if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
		SF_SetError("socket(AF_UNIX): %s", strerror(errno));
		return (-1);
	}
	Strlcpy(sockPath, _PATH_SOCKETDIR, sizeof(sockPath));
	Strlcat(sockPath, pw->pw_name, sizeof(sockPath));
	unaddr.sun_family = AF_UNIX;
	Strlcpy(unaddr.sun_path, sockPath, sizeof(unaddr.sun_path));
	socklen = SUN_LEN(&unaddr);
tryconn:
	if (connect(sock, (struct sockaddr *)&unaddr, socklen) == -1) {
		if (retry++ == 0) {
			if (CreateSpamCheckProcess(msg, sockPath, pw) == -1) {
				return (-1);
			}
			goto tryconn;
		} else {
			/* This should not happen. */
			if (retry > 10) {
				SF_SetError("Timeout waiting on spamcheck!");
				return (-1);
			}
			Debug("Waiting (%d/10) for process setup", retry);
			sleep(1);
		}
	}

	/* Write the message to the spamcheck socket. */
	Debug("Sending data (%d) to spamcheck", (int)msg->text_len);
	snprintf(wbuf, sizeof(wbuf), "%lu", (unsigned long)msg->text_len);
	if (write(sock, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
		SF_SetError("Write error (len)");
		goto fail;
	}
	for (wrote = 0; wrote < msg->text_len;) {
		ssize_t rv;

		rv = write(sock, &msg->text[wrote], msg->text_len - wrote);
		if (rv == -1) {
			if (errno == EINTR) {
				continue;
			}
			SF_SetError("Write error (msg)");
			goto fail;
		} else if (rv == 0) {
			SF_SetError("EOF from spamcheck");
			goto fail;
		}
		wrote += rv;
	}

	/* Read the result and required scores. */
	if (read(sock, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
		SF_SetError("Read error (score)");
		goto fail;
	}
	errno = 0;
	msg->status.score = (float)strtod(wbuf, &ep);
	if (*ep != '\0' || errno == ERANGE) {
		SF_SetError("Bad score: %s", wbuf);
		goto fail;
	}
	if (read(sock, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
		SF_SetError("Bad req_score value");
		goto fail;
	}
	errno = 0;
	msg->status.req_score = (float)strtod(wbuf, &ep);
	if (*ep != '\0' || errno == ERANGE) {
		SF_SetError("Bad req_score: %s", wbuf);
		goto fail;
	}

	/* Read the rewritten message from the spamcheck process. */
	if (read(sock, wbuf, sizeof(wbuf)) != sizeof(wbuf)) {
		SF_SetError("Bad len value");
		goto fail;
	}
	errno = 0;
	len = (size_t)strtoul(wbuf, &ep, 10);
	if (*ep != '\0' || errno == ERANGE) {
		SF_SetError("Bad len: %s", wbuf);
		goto fail;
	}
	msg->text_len = len;
	if ((msg->text = realloc(msg->text, len+1)) == NULL) {
		SF_SetError("Out of memory for rewritten message (%lu)",
		    (unsigned long)msg->text_len);
		msg->text_len = 0;
		goto fail;
	}
	for (nread = 0; nread < len;) {
		rv = read(sock, &msg->text[nread], len - nread);
		if (rv == -1) {
			if (errno == EINTR) {
				continue;
			}
			SF_SetError("Read error: %s", strerror(errno));
			goto fail;
		} else if (rv == 0) {
			syslog(LOG_ERR, "EOF from spamcheck process");
			break;
		}
		nread += rv;
	}
	msg->text[nread] = '\0';

	Debug("Spam score = %f (%f required), %lu bytes",
	    msg->status.score, msg->status.req_score,
	    (unsigned long)msg->text_len);
#if 0
wait:
	if (waitpid(pid, &status, 0) == -1) {
		if (errno == EINTR) {
			goto wait;
		}
		SF_SetError("wait: %s", strerror(errno));
		goto fail;
	}
	if (WIFSIGNALED(status)) {
		SF_SetError("signal %d", WTERMSIG(status));
		goto fail;
	} else if (WIFEXITED(status) && WEXITSTATUS(status)) {
		SF_SetError("failed (%d)", WEXITSTATUS(status));
		goto fail;
	}
#endif

	close(sock);
	return (0);
fail:
	close(sock);
	return (-1);
}

void
SPAM_Destroy(void)
{
	char path[1024];
	DIR *dir;
	struct dirent *dent;

	/* Unlink any socket that may have been left over. */
	if ((dir = opendir(_PATH_SOCKETDIR)) != NULL) {
		while ((dent = readdir(dir)) != NULL) {
			if (dent->d_name[0] == '.') {
				continue;
			}
			Strlcpy(path, _PATH_SOCKETDIR, sizeof(path));
			Strlcat(path, dent->d_name, sizeof(path));
			Debug("Removing stale socket: %s", path);
			unlink(path);
		}
		closedir(dir);
	} else {
		syslog(LOG_ERR, "%s: %s", _PATH_SOCKETDIR, strerror(errno));
	}
}

#endif /* HAVE_SA */
